package com.persagy.adm.diagram.core.line;

import cn.hutool.core.collection.CollUtil;
import com.persagy.adm.diagram.core.model.ConnectPoint;
import com.persagy.adm.diagram.core.model.DiagramNode;
import com.persagy.adm.diagram.core.model.EquipmentNode;
import com.persagy.adm.diagram.core.model.Line;
import com.persagy.adm.diagram.core.model.base.Container;
import com.persagy.adm.diagram.core.model.base.XY;
import com.persagy.adm.diagram.core.build.CalcContext;
import com.persagy.adm.diagram.core.model.template.MainPipe;
import com.persagy.adm.diagram.core.util.GeomUtil;
import org.locationtech.jts.geom.Point;

import java.util.*;

/**
 * 连线布局计算
 * @author zhaoyk
 */
public class LineLayoutManager {

	private CalcContext context;

	private List<Block> blocks = new ArrayList<>();

	private Map<Line, ConnectPoint[]> lines = new HashMap<>();

	public LineLayoutManager(CalcContext context) {
		this.context = context;
		initBlocks();
	}

	private void initBlocks(){
		for(Object o : context.getEquipMap().values()){
			if (o instanceof DiagramNode) {
				blocks.add(new Block((DiagramNode)o));
			}
		}
	}

	public List<XY> layout(Line line){
		ConnectPoint p1 = line.getFrom();
		ConnectPoint p2 = line.getTo();

		p1.layout(context);
		p2.layout(context);

		boolean exchange = false;
		//先计算点模型的终点
		if(p1.getEquipmentNode() == null) {
			if(p2.getEquipmentNode() == null) {
				throw new RuntimeException("do not support 2 main pipe connect");
			}

			ConnectPoint tmp = p2;
			p2 = p1;
			p1 = tmp;

			exchange = true;
		}

		if(checkLine(p1, p2)) {
			lines.put(line, new ConnectPoint[]{p1, p2});

			MainPipe passBy = findPassMainPipe(line);
			if(passBy != null) {
				ConnectPoint p3 = new ConnectPoint();
				p3.setHostObj(passBy);
				p3.setHostType(MainPipe.TYPE);

				List<XY> points1 = calcPath(p1, p3, false);
				List<XY> points2 = calcPath(p2, p3, true);

				List<XY> passByPoints = getPassByPoints(points1, points2, passBy);
				if(passByPoints != null){
					points1.addAll(passByPoints);
				}
				points1.addAll(points2);

				if(exchange) {
					Collections.reverse(points1);
				}
				return points1;
			} else {
				return calcPath(p1, p2, exchange);
			}
		} else {
			line.setFlag(Line.FLAG_DUPLICATE);
			return buildDuplicateMark(p1, p2);
		}
	}

	private MainPipe findPassMainPipe(Line line){
		EquipmentNode fromNode = line.getFrom().getEquipmentNode();
		EquipmentNode toNode = line.getTo().getEquipmentNode();

		if(fromNode == null || toNode == null){
			return null;
		}

		List<MainPipe> mainPipes = context.getTemplate().getMainPipes();
		if (CollUtil.isNotEmpty(mainPipes)) {
			for(MainPipe mp : mainPipes) {
				if (!mp.isBindEquipment() && mp.getPassbyRels() != null && mp.getPassbyRels().contains(line.getRelType())) {
					boolean flag1 = false, flag2 = false;
					List<Container> relCons = mp.getRelatedContainerList();
					if(relCons != null) {
						for (Container con : relCons) {
							if(!flag1 && con.isParentOf(fromNode)){
								flag1 = true;
							}
							if(!flag2 && con.isParentOf(toNode)){
								flag2 = true;
							}
							if(flag1 && flag2){
								break;
							}
						}
					}
					if (flag1 && flag2) {
						return mp;
					}
				}
			}
		}
		return null;
	}

	private List<XY> getPassByPoints(List<XY> part1, List<XY> part2, MainPipe passBy){
		XY p1 = part1.get(part1.size() - 1);
		XY p2 = part2.get(0);
		for (List<XY> line : passBy.getLocationPath()) {
			int idx1 = -1, idx2 = -1;
			XY pre = null;
			int iter = 0;
			for(XY point : line) {
				if(pre != null) {
					if(idx1 == -1 && onLine(pre, point, p1)){
						idx1 = iter;
					}
					if(idx2 == -1 && onLine(pre, point, p2)){
						idx2 = iter;
					}
					if(idx1 >= 0 && idx2 >= 0){
						break;
					}

					iter++;
				}
				pre = point;
			}

			if(idx1 >= 0 && idx2 >= 0 && idx1 != idx2) {
				List<XY> passByPoints = new LinkedList<>();
				boolean r = idx1 > idx2;
				for(int i = (!r ? idx1 : idx2); i < (!r ? idx2 : idx1); i++){
					passByPoints.add(line.get(i + 1));
				}
				if(r){
					Collections.reverse(passByPoints);
				}
				return passByPoints;
			}
		}
		return null;
	}

	private boolean onLine(XY lineEnd1, XY lineEnd2, XY point) {
		int begin, end;
		if(lineEnd1.x == lineEnd2.x){
			if(point.x == lineEnd1.x){
				begin = Math.min(lineEnd1.y, lineEnd2.y);
				end = Math.max(lineEnd1.y, lineEnd2.y);
				return point.y >= begin && point.y <= end;
			}
		} else {
			if(point.y == lineEnd1.y){
				begin = Math.min(lineEnd1.x, lineEnd2.x);
				end = Math.max(lineEnd1.x, lineEnd2.x);
				return point.x >= begin && point.x <= end;
			}
		}
		return false;
	}

	private List<XY> calcPath(ConnectPoint p1, ConnectPoint p2, boolean exchange){
		LineEnd[] arr = buildLineEnd(p1, p2);
		List<XY> points = new PathBuilder(arr[0], arr[1], p1).calcPath(blocks);
		finishPath(points, p1, p2, exchange);
		return points;
	}

	private void finishPath(List<XY> points, ConnectPoint p1, ConnectPoint p2, boolean exchange){
		if (p1.getLocation() != null) {
			points.add(0, p1.getLocation());
		}
		if (p2.getLocation() != null) {
			points.add(p2.getLocation());
		}

		if (exchange) {
			Collections.reverse(points);
		}
	}

	private LineEnd[] buildLineEnd(ConnectPoint p1, ConnectPoint p2) {
		LineEnd[] arr = new LineEnd[2];

		arr[0] = new LineEnd();
		prepareAnchorLineEnd(arr[0], p1);

		if(p2.getHostObj() instanceof EquipmentNode) {
			arr[1] = new LineEnd();
			prepareAnchorLineEnd(arr[1], p2);
		} else if (p2.getHostObj() instanceof MainPipe) {
			Point point1 = GeomUtil.getPoint(p1.getLocation());
			MainPipe mp = (MainPipe) p2.getHostObj();
			XY[] nearLine = null;
			double min = Double.MAX_VALUE;
			for(List<XY> list : mp.getLocationPath()) {
				XY pre = null;
				for(XY xy : list) {
					if(pre != null) {
						//复制坐标
						XY[] line = new XY[]{pre.copy(), xy.copy()};
						//起点终点排序
						LineEnd.orderLine(line);

						//线需要留头
						if(LineEnd.isHorizontal(line)){
							if(line[1].x - line[0].x > LineEnd.offset * 2){
								line[0].x += LineEnd.offset;
								line[1].x -= LineEnd.offset;
							}
						} else {
							if(line[1].y - line[0].y > LineEnd.offset * 2){
								line[0].y += LineEnd.offset;
								line[1].y -= LineEnd.offset;
							}
						}
						//计算点和线段的最短距离
						double distance = point1.distance(GeomUtil.getLine(line));
						if(distance < min){
							min = distance;
							nearLine = line;
						}
					}
					pre = xy;
				}
			}
			arr[1] = new LineEnd();
			arr[1].setLine(nearLine);
		}

		return arr;
	}

	private void prepareAnchorLineEnd(LineEnd lineEnd, ConnectPoint connectPoint) {
		lineEnd.setPoint(anchorEmerge(connectPoint));
		lineEnd.setAnchorSide(getAnchorSide(connectPoint));
	}

	/**
	 * 锚点出线
	 */
	private XY anchorEmerge(ConnectPoint point){
		XY xy = new XY(point.getLocation());
		switch (point.getAnchorCode().charAt(0)) {
			case 'T':
				xy.y -= LineEnd.offset;
				break;
			case 'B':
				xy.y += LineEnd.offset;
				break;
			case 'L':
				xy.x -= LineEnd.offset;
				break;
			case 'R':
				xy.x += LineEnd.offset;
				break;
			default:
		}
		return xy;
	}

	/**
	 * 获取锚点所在的节点边路径
	 */
	private XY[] getAnchorSide(ConnectPoint point){
		EquipmentNode en = point.getEquipmentNode();
		XY loc = en.locationToRoot();
		XY size = en.getSize();
		switch (point.getAnchorCode().charAt(0)) {
			case 'T':
				return new XY[] {loc, new XY(loc.x + size.x, loc.y)};
			case 'B':
				return new XY[] {new XY(loc.x, loc.y + size.y), new XY(loc.x + size.x, loc.y + size.y)};
			case 'L':
				return new XY[] {loc, new XY(loc.x, loc.y + size.y)};
			case 'R':
				return new XY[] {new XY(loc.x + size.x, loc.y), new XY(loc.x + size.x, loc.y + size.y)};
			default:
		}
		return null;
	}

	private boolean checkLine(ConnectPoint p1, ConnectPoint p2){
		for(ConnectPoint[] points : lines.values()) {
			if(p1.sameWith(points[0]) && p2.sameWith(points[1])){
				return false;
			}
		}
		return true;
	}

	private List<XY> buildDuplicateMark(ConnectPoint p1, ConnectPoint p2){
		ArrayList<XY> markPath = new ArrayList<>();
		markPath.add(p1.getLocation());
		if(p2.getEquipmentNode() != null){
			markPath.add(p2.getLocation());
		} else {
			markPath.add(p2.getMainPipe().getLocationPath().get(0).get(0));
		}
		return markPath;
	}

	public List<Block> getBlocks() {
		return blocks;
	}

	public CalcContext getContext() {
		return context;
	}

}
